import inspect
import os
from random import randint
from uuid import uuid4

import pytest

from pyqalx.config import BotConfig
from pyqalx.core import QalxNoGUIDError
from pyqalx.core.entity import QalxEntity
from pyqalx.core.errors import QalxError, QalxAPIResponseError, QalxEntityNotFound, QalxMultipleEntityReturned, \
    QalxNoQueueFoundByName
from pyqalx.core.group import Group
from pyqalx.core.item import Item
from pyqalx.core.qalx import entity_by_string
from pyqalx.core.queue import Queue
from pyqalx.core.set import Set


@pytest.mark.parametrize("request_type", ['get', 'post', 'patch', 'delete'])
@pytest.mark.parametrize("endpoint", ['item', 'set', 'group', 'queue'])
def test_process_api_request(qalx, mocker, request_type, endpoint):
    mock_transport_function = mocker.patch("pyqalx.core.qalx.PyQalxAPI." + request_type)
    mock_transport_function.return_value = (True, {"guid": uuid4()})
    qalx._process_api_request(request_type, endpoint, thing="stuff")
    mock_transport_function.assert_called_once_with(endpoint, json={"meta": {},
                                                                    "thing": "stuff"})
    pass


def test_process_api_request_error(qalx):
    with pytest.raises(Exception):
        qalx._process_api_request("foo", "here", thing="stuff")


@pytest.mark.parametrize("request_type", ['get', 'post', 'patch', 'delete'])
@pytest.mark.parametrize("endpoint", ['item', 'set', 'group', 'queue'])
def test_process_api_request_non_json_error(qalx, mocker, request_type, endpoint):
    mock_transport_function = mocker.patch("pyqalx.core.qalx.PyQalxAPI." + request_type)
    mock_transport_function.return_value = (True, {"guid": uuid4()})
    with pytest.raises(QalxError):
        qalx._process_api_request(request_type, endpoint, thing=uuid4())


@pytest.mark.parametrize("request_type", ['get', 'post', 'patch', 'delete'])
@pytest.mark.parametrize("endpoint", ['item', 'set', 'group', 'queue'])
def test_process_api_request_http_error(qalx, mocker, request_type, endpoint):
    mock_transport_function = mocker.patch("pyqalx.core.qalx.PyQalxAPI." + request_type)
    mock_transport_function.return_value = (False, {"I'm a": "complete HTTP failure"})
    with pytest.raises(QalxAPIResponseError) as e:
        qalx._process_api_request(request_type, endpoint, thing="stuff")
        assert "I'm a: complete HTTP failure" in str(e)


def test_entity_by_string(mocker):
    class EntityA(QalxEntity):
        entity_type = "A"

    class EntityB(QalxEntity):
        entity_type = "B"

    class EntityCButNamedB(QalxEntity):
        entity_type = "B"

    mock_function = mocker.patch('pyqalx.core.qalx.getmembers')
    mock_function.return_value = [("EntityA", EntityA)]
    assert entity_by_string("A") == EntityA

    mock_function.return_value = []
    with pytest.raises(QalxEntityNotFound):
        entity_by_string("B")

    mock_function.return_value = [('EntityB', EntityB), ("EntityCButNamedB", EntityCButNamedB)]
    with pytest.raises(QalxMultipleEntityReturned):
        entity_by_string("B")


@pytest.mark.parametrize("endpoint", ['item', 'set', 'group', 'queue'])
def test_entities(endpoint):
    assert QalxEntity in inspect.getmro(entity_by_string(endpoint))


@pytest.mark.parametrize("entity_type", ['item', 'set', 'group', 'queue'])
def test_get_entity_by_guid(qalx, mocker, entity_type):
    guid = uuid4().hex
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    mocked_process_api_request.return_value = {"guid": guid}
    mocked_entity_by_string = mocker.patch("pyqalx.core.qalx.entity_by_string")
    mocked_entity_by_string.return_value = QalxEntity
    entity = qalx.get_entity_by_guid(entity_type, guid)
    assert entity['guid'] == guid
    endpoint = f"{entity_type}/{guid}"
    mocked_process_api_request.assert_called_with('get', endpoint)


@pytest.mark.parametrize("entity_type", ['item', 'set', 'group', 'queue'])
def test_archive_entity_by_guid(qalx, mocker, entity_type):
    guid = uuid4().hex
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    mocked_process_api_request.return_value = {"guid": guid, "info": {"archived": True}}
    mocked_entity_by_string = mocker.patch("pyqalx.core.qalx.entity_by_string")
    mocked_entity_by_string.return_value = QalxEntity
    entity = qalx.archive_entity_by_guid(entity_type, guid)
    assert entity['guid'] == guid
    endpoint = f"{entity_type}/{guid}/archive"
    mocked_process_api_request.assert_called_with('patch', endpoint)


@pytest.mark.parametrize("entity_type", ['item', 'set', 'group', 'queue'])
def test_save_entity(qalx, mocker, entity_type):
    Entity = entity_by_string(entity_type)
    guid = uuid4()
    entity_structure = {"guid": guid, "info": {"archived": False}, "data": {"thing": "stuff"}}
    entity = Entity({})
    with pytest.raises(QalxNoGUIDError):
        qalx.save_entity(entity)
    entity = Entity(entity_structure)
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    endpoint = f"{entity_type}/{guid}"
    mocked_process_api_request.return_value(entity_structure)
    qalx.save_entity(entity)
    mocked_process_api_request.assert_called_with('patch', endpoint, data=entity_structure['data'])


def test_add_item_data(qalx, mocker):
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    guid = uuid4()
    item_structure = {"guid": guid, "info": {"archived": False}, "data": {"thing": "stuff"}}
    mocked_process_api_request.return_value = item_structure
    item1 = qalx.add_item_data(data=item_structure['data'])
    mocked_process_api_request.assert_called_with('post', 'item', data=item_structure['data'], meta=None)
    assert isinstance(item1, Item)
    meta = {"some": "meta data"}
    item2 = qalx.add_item_data(data=item_structure['data'], meta=meta)
    mocked_process_api_request.assert_called_with('post', 'item', data=item_structure['data'], meta=meta)
    assert isinstance(item2, Item)
    with pytest.raises(QalxError):
        qalx.add_item_data(data='data', meta=meta)


def test_add_item_file(qalx, mocker):
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    guid = uuid4()
    item_structure = {"guid": guid, "info": {"archived": False}, "file": {"url": "s3://thefile"}}
    mocked_process_api_request.return_value = item_structure
    file_path = "/path/to/a/file"
    item1 = qalx.add_item_file(file_path=file_path)
    mocked_process_api_request.assert_called_with('post', 'item', file_path=file_path, data={}, meta=None)
    assert isinstance(item1, Item)


def test_get_item_file(qalx, mocker):
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    guid = uuid4().hex
    item_structure = {"guid": guid, "info": {"archived": False},
                      "file": {"url": "s3://thefile", "filename": "some_file.extension"}}
    mocked_process_api_request.return_value = item_structure
    item1 = qalx.get_entity_by_guid('item', guid)
    mocked_process_api_request.assert_called_with('get', 'item/'+ guid)
    mocked_requests = mocker.patch('pyqalx.core.item.requests')
    assert isinstance(item1, Item)

    some_bytes = b'some bytes'
    mock_response = mocker.MagicMock()
    mocked_requests.get.return_value = mock_response
    mock_response.content = some_bytes
    mock_response.ok = True
    file_bytes = item1.read_file()
    mocked_requests.get.assert_called_with(url=item_structure['file']['url'])
    assert file_bytes == some_bytes
    fake_path = "/a/path/"
    mocked_open = mocker.mock_open()
    m = mocker.patch('pyqalx.core.item.open',mocked_open, create=True)
    item1.save_file(fake_path)
    full_fake_path = os.path.join(fake_path, item_structure['file']['filename'])
    mocked_open.assert_called_with(full_fake_path, 'wb')


def test_add_set(qalx, mocker):
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    guid1 = uuid4().hex
    guid2 = uuid4().hex
    item1_structure = {"guid": guid1, "info": {"archived": False}, "data": {"thing": "stuff"}}
    item2_structure = {"guid": guid2, "info": {"archived": False}, "data": {"thing": "stuff"}}
    mocked_process_api_request.return_value = {"guid": uuid4(), "info": {"some": "info"}, "item_guids": [guid1, guid2]}
    set = qalx.add_set(items=[Item(item1_structure), Item(item2_structure)])
    mocked_process_api_request.assert_called_with('post', 'set', item_guids=[guid1, guid2], meta=None)
    assert isinstance(set, Set)


def test_add_group(qalx, mocker):
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    guid1 = uuid4().hex
    guid2 = uuid4().hex
    set1_structure = {"guid": guid1, "info": {"archived": False}, "item_guids": ["item1", "item2"]}
    set2_structure = {"guid": guid2, "info": {"archived": False}, "item_guids": ["item3", "item4"]}
    mocked_process_api_request.return_value = {"guid": uuid4(), "info": {"some": "info"}, "set_guids": [guid1, guid2]}
    group = qalx.add_group(sets=[Set(set1_structure), Set(set2_structure)])
    mocked_process_api_request.assert_called_with('post', 'group', set_guids=[guid1, guid2], meta=None)
    assert isinstance(group, Group)


def test_add_queue(qalx, mocker):
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    queue = qalx.add_queue("unittest queue", meta={"some": "stuff"})
    mocked_process_api_request.assert_called_with('post', 'queue', parameters=qalx._queue_params,
                                                  meta={"some": "stuff", "queue_name": "unittest queue"})
    assert isinstance(queue, Queue)
    queue = qalx.add_queue("unittest queue")
    mocked_process_api_request.assert_called_with('post', 'queue', parameters=qalx._queue_params,
                                                  meta={"queue_name": "unittest queue"})
    assert isinstance(queue, Queue)


@pytest.mark.parametrize("entity_type", ['item', 'set', 'group', 'queue'])
def test_list_entities(qalx, mocker, entity_type):
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    params = dict(sort=None, skip=randint(0, 100), limit=randint(1, 500))
    entities = qalx.list_entities(entity_type, **params)
    mocked_process_api_request.assert_called_with("get", entity_type, **params)


def test_get_queue_messages(qalx, mocker):
    mock_queue = mocker.MagicMock()
    with pytest.raises(QalxError):
        qalx.get_queue_messages(mock_queue)


def test_get_queue_by_name(qalx, mocker):
    list_method = mocker.patch.object(qalx, 'list_entities')
    mock_queues = [mocker.MagicMock(), mocker.MagicMock()]
    list_method.return_value = mock_queues
    with pytest.raises(QalxMultipleEntityReturned):
        qalx.get_queue_by_name("unittest queue")

    queue = Queue({"guid": uuid4()})
    list_method.return_value = [queue]
    get_by_guid_method = mocker.patch.object(qalx, 'get_entity_by_guid')
    q = qalx.get_queue_by_name("unittest queue")
    get_by_guid_method.assert_called_with('queue', queue['guid'])


def test_get_or_create_queue(qalx, mocker):
    get_by_name_method = mocker.patch.object(qalx, 'get_queue_by_name')
    get_by_name_method.return_value = Queue({"meta": {"name": "unittest queue"}})
    q = qalx.get_or_create_queue("unittest queue")
    get_by_name_method.assert_called_with("unittest queue")
    assert isinstance(q, Queue)
    get_by_name_method = mocker.patch.object(qalx, 'get_queue_by_name')
    get_by_name_method.side_effect = QalxNoQueueFoundByName
    create_queue_method = mocker.patch.object(qalx, 'add_queue')
    q = qalx.get_or_create_queue("unittest new queue", meta={"some": "meta"})
    create_queue_method.assert_called_with("unittest new queue", {"some": "meta"})


def test_add_bot(qalx, mocker):
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    conf = {"Some": "config"}
    bot_config = BotConfig(conf)
    bot = qalx.add_bot(bot_config)
    mocked_process_api_request.assert_called_with('post', 'bot', config=conf, host=qalx._host_info, meta=None)


def test_update_bot_status(qalx, mocker):
    mocked_process_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')
    bot = qalx.update_bot_status("1234", "unittest")
    mocked_process_api_request.assert_called_with('patch', 'bot/1234', status="unittest")




import pytest

from pyqalx.config import Config, BotConfig, UserConfig


def test_base_config_class():
    config = Config()
    assert not config.key_prefix
    assert not config.filename


@pytest.mark.parametrize("config_type,thing_returns", [
    (BotConfig, "A BOT THING"),
    (UserConfig, "A USER THING")
])
def test_envvar_config(fake_environ, config_type, thing_returns):
    config = config_type()
    assert config['THING'] == thing_returns
    assert "SOME_OTHER_THING" not in config


@pytest.mark.parametrize("config_type", [BotConfig, UserConfig])
def test_botsinifile_config(mocker, config_type):
    mocked_path = mocker.patch('pyqalx.config.os.path.exists')
    mocked_path.return_value = True
    mo = mocker.mock_open(read_data="[default]\nSTUFF=SOME STUFF")
    mocker.patch('builtins.open', mo)

    config = config_type()
    config.from_inifile()
    mo.assert_called_once_with(config.config_path)
    assert config['STUFF'] == "SOME STUFF"



import json

import boto3

from pyqalx.core.entity import QalxEntity
from pyqalx.core.errors import QalxQueueError
from pyqalx.core.group import Group
from pyqalx.core.item import Item
from pyqalx.core.set import Set


class QueueResponse:
    """a response from a remote queue"""

    def __init__(self, response):
        """make new response

        v0.1 - doesn't do much yet
        :param response: response from queue
        """
        self._raw_response = response


class QueueMessage:
    """a message frm the queue

    attributes:
        body: message body
        receipt_handle: some information to return to the message broker to confirm receipt
    """

    def __init__(self, message):
        """

        :param message: the full raw message from the queue
        """
        self._raw_message = message
        self.body = json.loads(message['Body'])
        self.receipt_handle = message['ReceiptHandle']


class Queue(QalxEntity):
    """QalxEntity with entity_type Queue

    attributes:
        broker_client: an authenticated client to communicate with the remote message broker
    """
    entity_type = 'queue'

    def _stringify(self, message: dict) -> str:
        return json.dumps(message)

    def __init__(self, *args, **kwargs):
        super(Queue, self).__init__(*args, **kwargs)

        if self.get('info') and ('credentials' in self['info']):
            self.broker_client = boto3.client('sqs',
                                              region_name="eu-west-2",  # TODO: make this configurable
                                              aws_access_key_id=self['info']['credentials']['ACCESS_KEY_ID'],
                                              aws_secret_access_key=self['info']['credentials']['SECRET_ACCESS_KEY'],
                                              aws_session_token=self['info']['credentials']['SESSION_TOKEN'])
        else:
            self.broker_client = None

    def submit_sets_from_group(self, group):
        """

        :param group: a group
        :type group: pyqalx.core.Group
        :returns: pyqalx.core.queue.QueueResponse
        """
        entries = []
        for n, set_guid in enumerate(group['set_guids']):
            entries.append(
                {
                    "Id": str(n),
                    'MessageBody': json.dumps({
                        "entity_type": "set",
                        "entity_guid": set_guid,
                        "parent_group_guid": group['guid']
                    })
                })
        response = self.broker_client.send_message_batch(
            QueueUrl=self['info']['queue_url'],
            Entries=entries
        )
        return QueueResponse(response)

    def get_messages(self, max_num_msg, visibility, waittime):
        """get messages from the queue

        :param max_num_msg: maximum number to retrieve
        :param visibility: time in seconds until the message becomes visible again on the queue
        :param waittime: time to wait for a message
        :return: list of pyqalx.core.queue.QueueMessage
        """
        responses = self.broker_client.receive_message(
            QueueUrl=self['info']['queue_url'],
            MaxNumberOfMessages=max_num_msg,
            VisibilityTimeout=visibility,
            WaitTimeSeconds=waittime,
        )
        if not responses.get("Messages"):
            return []
        else:
            return [QueueMessage(response) for response in responses['Messages']]

    def delete_message(self, message):
        """remove message from the queue

        :param message:
        :type message QalxMessage
        :return:
        """
        self.broker_client.delete_message(
            QueueUrl=self['info']['queue_url'],
            ReceiptHandle=message.receipt_handle
        )

    def submit_entity(self, entity):
        """put a qalx entity on the queue

        :param entity: entity to add to the queue
        :type entity: pyqalx.core.entity.QalxEntity
        :return: pyqalx.core.queue.QueueResponse
        """
        if not type(entity) in [Item, Set, Group]:
            raise QalxQueueError("Only Item, Set and Group entities can be submitted to the queue.")
        response = self.broker_client.send_message(QueueUrl=self['info']['queue_url'],
                                                   MessageBody=self._stringify(
                                                       {"entity_type": entity.entity_type,
                                                        "entity_guid": entity['guid']
                                                        }))
        return QueueResponse(response)



import os
try:
    from json.decoder import JSONDecodeError
except ImportError:
    # JSONDecodeError is for Python3 only
    JSONDecodeError = ValueError

import requests


class PyQalxAPIException(Exception):
    """
    A generic error for PyQalxAPI
    """
    pass


class BasePyQalxAPI(object):
    def __init__(self, config):
        self.base_url = config.get('BASE_URL', 'https://api.qalx.io/')
        self.token = config['TOKEN']
        self._last_response = None

    def _build_request(self, url, method, include_auth_headers=True, **kwargs):
        """
        Handles making the request
        :param url: The url to make the request to
        :param include_auth_headers: Should we include the auth headers
        :param kwargs: kwargs to pass through to the request
        :return: response
        """
        if include_auth_headers:
            # We might not include the headers if we are PUTing to S3
            headers = kwargs.get('headers', {})
            # Don't overwrite existing headers
            headers.update({'Authorization': 'Token {token}'.format(
                token=self.token)})
            kwargs['headers'] = headers
        try:
            resp = requests.request(url=url,
                                    method=method,
                                    **kwargs)
            self._last_response = resp
        except requests.exceptions.RequestException as exc:
            # RequestException: The base exception that `requests` can raise.
            #                   Catch everything and return to the client
            resp = exc
        return self._build_response(resp)

    def _build_non_safe_request(self,
                                method,
                                endpoint,
                                upload,
                                delete_file=False,
                                **kwargs):
        """
        Helper method for performing non safe requests (POST, PUT, PATCH etc).
        Handles the possibility of a user trying to create or update an item
        that also has a file
        :param method: The HTTP method
        :param endpoint: The endpoint you want to query: 'item`
        :param upload: If you are specifying file data should this method
                       attempt to upload the file to S3
        :param kwargs: Any data you want to pass to the request.
                      Use `json` key for data you want to post. Anything else
                      will be passed to `requests` as kwargs
                      json={'data': {'some': 'data'}, 'meta': {'some': 'meta'}}
        :return: response
        """
        url = self._get_url(endpoint)
        file_path = kwargs.pop('file_path', None)
        data = kwargs.get('json', {})

        if file_path is not None and delete_file:
            raise PyQalxAPIException('`file` and `delete_file` have both '
                                     'been specified. Please specify only one')
        if file_path is not None:
            if os.path.isfile(file_path):
                file_path = os.path.abspath(file_path)
                filename = os.path.basename(file_path)
                data['file'] = {'name': filename}
            else:
                raise PyQalxAPIException('Invalid file "{0}"'.format(file_path))  # noqa
        elif delete_file:
            data['file'] = {}

        kwargs['json'] = data
        resp_ok, data = self._build_request(url=url,
                                            method=method,
                                            **kwargs)
        if upload and resp_ok and file_path is not None:
            # Only upload if everything was OK with the item creation
            # and the user wants to upload the file automatically.
            s3_resp_ok, s3_data = self._upload_to_s3(data, file_path)
            if s3_resp_ok is False:
                # The S3 upload failed, return the failed S3 response in place
                # of the response from the api.  Any implementing client will
                # have to handle tidying up any incorrect data.
                resp_ok = s3_resp_ok
                data = s3_data
            else:
                # Always pop the `put_url` off if it's a success
                data['file'].pop('put_url')
        return resp_ok, data

    def _build_response(self, resp):
        """
        Builds the response that gets sent back to the client
        :param resp: The response from `requests` or an instance
                     of `RequestException`
        :return: tuple of `(response_ok:bool, response data:dict)`
        """
        # If it doesn't have `ok` then a `requests` exception was raised
        is_ok = getattr(resp, 'ok', False)

        try:
            data = resp.json()
        except (AttributeError, JSONDecodeError):
            # AttributeError: `resp` is an instance of RequestException
            # JSONDecodeError: `resp` is a normal response but has no data.
            #                   i.e. a 500 or a `delete` response
            data = []
        if is_ok is False:
            # If either of these are missing then a `requests` exception was
            # raised
            status_code = getattr(resp, 'status_code', '')
            reason = getattr(resp, 'reason', resp)
            data = {
                'status_code': status_code,
                'reason': reason,
                'errors': data
            }
        return is_ok, data

    def _get_url(self, endpoint):
        """
        Builds the URL for the request from the base_url and the endpoint
        :param endpoint: The endpoint you want to query: 'item/<item_guid>'
        :return: url
        """
        url = '{base_url}{endpoint}'.format(
            base_url=self.base_url,
            endpoint=endpoint.rstrip('/').lstrip('/')
        )
        return url

    def _upload_to_s3(self, data, file_path):
        """
        Given a response object with a file url in will attempt to
        PUT the file onto S3
        :param data: The data from the api
        :param file_path: The path to the file to upload
        :return: response
        """
        url = data['file']['put_url']
        return self._build_request(url=url,
                                   method='PUT',
                                   include_auth_headers=False,
                                   data=open(file_path, 'rb').read())



class QalxEntity(dict):
    entity_type: str

    def __init__(self, pyqalxapi_dict):
        """Base class for qalx entities_response.

        QalxEntity children need to be populated with either a `requests.models.Response` which is the type returned
        by the methods on `pyqalxapi.api.PyQalxAPI` or with a `dict`.

        Entities will behave exactly like a `dict`. For example:

        >>> class AnEntity(QalxEntity):
        ...     pass
        >>> c = AnEntity({"guid":"123456789", "info":{"some":"info"}})
        >>> c['guid']
        '123456789'

        :param pyqalxapi_dict: a 'dict' representing a qalx entity object to populate the entity
        :type pyqalxapi_dict: dict
        """

        super().__init__(pyqalxapi_dict)

    # TODO: tricky for nested dicts
    #     self._dirty_keys = []
    #
    # def __setitem__(self, key, value):
    #     self._dirty_keys.append(key)
    #
    # @property
    # def keys_to_save(self):
    #     return {k: self[k] for k in self._dirty_keys}
    #
    # @property
    # def is_dirty(self):
    #     return self._dirty_keys is not []

    def __str__(self):
        return f"[{self.entity_name}] {self['guid']}"



import json
from multiprocessing import Manager

import pytest

from pyqalx import Bot
from tests.model_entities import queue_response, entities_response, bot_post


def bot_processes(bot, jobs, expected_output):
    """
    Helper function for handling the bot processes
    :param bot: The bot instance
    :param jobs: An instance of Manager().list().  Used to keep track of all
                 the jobs across multiple processes
    :param expected_output: An instance of Manager().list().  Used to keep
                            track of all outputs for the jobs across multiple
                            processes
    :return: bot, jobs, expected_output
    """
    @bot.initialisation
    def _initialisation(user_qalx):
        jobs.append('init_complete')
        return True

    @bot.begin
    def _begin(_job):
        val = "begin"
        _job.add_step_result(True, {"test": val})
        jobs.append(_job.last_step_result)
        expected_output.append(val)

    @bot.preload
    def _preload(_job):
        val = 'preload'
        _job.add_step_result(True, {"test": val})
        jobs.append(_job.last_step_result)
        expected_output.append(val)

    @bot.onload
    def _onload(_job):
        val = 'onload'
        _job.add_step_result(True, {"test": val})
        jobs.append(_job.last_step_result)
        expected_output.append(val)

    @bot.preprocess
    def _preprocess(_job):
        val = 'preprocess'
        _job.add_step_result(True, {"test": val})
        jobs.append(_job.last_step_result)
        expected_output.append(val)

    @bot.process
    def _process(_job):
        val = 'process'
        _job.add_step_result(True, {"test": val})
        jobs.append(_job.last_step_result)
        expected_output.append(val)

    @bot.precompletion
    def _precompletion(_job):
        val = 'precompletion'
        _job.add_step_result(True, {"test": val})
        jobs.append(_job.last_step_result)
        expected_output.append(val)

    @bot.postprocess
    def _postprocess(_job):
        val = 'postprocess'
        _job.add_step_result(True, {"test": val})
        _job.stop = True
        jobs.append(_job.last_step_result)
        expected_output.append(val)

    @bot.ontermination
    def _ontermination(_job):
        val = 'termination'
        _job.add_step_result(True, {"test": val})
        jobs.append(_job.last_step_result)
        expected_output.append(val)

    @bot.onwait
    def _onwait(_job):
        val = 'onwait'
        _job.add_step_result(True, {"test": val})
        jobs.append(_job.last_step_result)
        expected_output.append(val)
    return bot, jobs, expected_output

@pytest.mark.xfail
def test_bot_runs_all_functions(mocker, qalx_class):
    """
    Tests that a bot runs through all the expected functions and processes
    them accordingly
    """
    # No response from status update
    status_update = {}
    get_queues = {
        'data': [queue_response],
    }

    get_item = entities_response['item'][0]
    mocked_api_request = mocker.patch(
        'pyqalx.core.qalx.Qalx._process_api_request')
    mocked_api_request.side_effect = [
        # The first call creates a bot
        bot_post,
        status_update,
        get_queues,
        queue_response,
        entities_response['item'][0],
    ]
    mocked_boto3 = mocker.patch('pyqalx.core.queue.boto3')
    mocked_boto3.client.return_value.receive_message.side_effect = [
        {'Messages': [{'Body': json.dumps({
            'entity_type': 'item',
            'entity_guid': get_item['guid']
        }),
            'ReceiptHandle': 'receipt handle'}]}
    ]
    bot = Bot(bot_name='test_bot',
              queue_name='test_queue',
              skip_ini=True,
              qalx_class=qalx_class)

    with Manager() as manager:
        # Use a manager for a shared state across all processes
        jobs = manager.list()
        expected_output = manager.list()

        bot, jobs, expected_output = bot_processes(bot, jobs, expected_output)

        bot.start(processes=1)
        assert jobs[0] == 'init_complete'

        for cnt, job in enumerate(jobs[1:]):
            assert job.result_data == {'test': expected_output[cnt]}
            # This test case should never wait
            assert job.result_data != {'test': 'onwait'}, \
                "BotProcess called `onwait` function when it shouldn't have"

@pytest.mark.xfail
@pytest.mark.slow
def test_bot_runs_all_functions_wait_response(mocker, qalx_class):
    """
    Tests that a bot runs through all the expected functions and processes
    them accordingly.  This also tests that the wait response works in case
    we don't get a message from SQS.
    This will take a while to complete due to the
    `sleep(wait_time * spread_factor)`
    """
    # No response from status update
    status_update = {}
    get_queues = {
        'data': [queue_response],
    }

    get_item = entities_response['item'][0]
    mocked_api_request = mocker.patch(
        'pyqalx.core.qalx.Qalx._process_api_request')
    mocked_api_request.side_effect = [
        # The first call creates a bot
        bot_post,
        status_update,
        get_queues,
        queue_response,
        entities_response['item'][0],
    ]
    mocked_boto3 = mocker.patch('pyqalx.core.queue.boto3')
    mocked_boto3.client.return_value.receive_message.side_effect = [
        {},
        {'Messages': [{'Body': json.dumps({
            'entity_type': 'item',
            'entity_guid': get_item['guid']
        }),
            'ReceiptHandle': 'receipt handle'}]}
    ]

    bot = Bot(bot_name='test_bot',
              queue_name='test_queue',
              skip_ini=True,
              qalx_class=qalx_class)

    with Manager() as manager:
        # Use a manager for a shared state across all processes
        jobs = manager.list()
        expected_output = manager.list()

        bot, jobs, expected_output = bot_processes(bot, jobs, expected_output)

        bot.start(processes=1)
        assert jobs[0] == 'init_complete'

        for cnt, job in enumerate(jobs[1:]):
            assert job.result_data == {'test': expected_output[cnt]}


@pytest.mark.slow
@pytest.mark.xfail
def test_bot_runs_all_functions_kill_after(mocker,
                                           qalx_class):
    """
    Tests that a bot runs through all the expected functions and processes
    them accordingly - killing the process if we have attempted to get messages
    more than `KILL_AFTER` times.  This also tests reading from the ini file.
    This will take a while to complete due to the
    `sleep(wait_time * spread_factor)`
    """
    # No response from status update
    status_update = {}
    get_queues = {
        'data': [queue_response],
    }

    get_item = entities_response['item'][0]
    mocked_api_request = mocker.patch('pyqalx.core.qalx.Qalx._process_api_request')  # noqa
    mocked_api_request.side_effect = [
        # The first call creates a bot
        bot_post,
        status_update,
        get_queues,
        queue_response,
        entities_response['item'][0],
    ]
    mocked_boto3 = mocker.patch('pyqalx.core.queue.boto3')
    mocked_boto3.client.return_value.receive_message.side_effect = [
        # This is key to the `KILL_AFTER` test.  We will kill after two
        # attempts.  Meaning that the actual message will never get read
        {},
        {},
        # This will never get read
        {'Messages': [{'Body': json.dumps({
            'entity_type': 'item',
            'entity_guid': get_item['guid']
        }),
            'ReceiptHandle': 'receipt handle'}]}
    ]
    # Mock the config file that is normally stored on the filesystem
    mocked_path = mocker.patch('pyqalx.config.os.path.exists')
    mocked_path.return_value = True
    mo = mocker.mock_open(read_data="[default]\nKILL_AFTER=2")
    mocker.patch('pyqalx.config.open', mo)

    bot = Bot(bot_name='test_bot',
              queue_name='test_queue',
              skip_ini=False,
              qalx_class=qalx_class)

    with Manager() as manager:
        # Use a manager for a shared state across all processes
        jobs = manager.list()
        expected_output = manager.list()

        bot, jobs, expected_output = bot_processes(bot, jobs, expected_output)

        bot.start(processes=1)
        assert jobs[0] == 'init_complete'

        for cnt, job in enumerate(jobs[1:]):
            assert job.result_data == {'test': expected_output[cnt]}
            for func_not_run in ['preload', 'onload', 'preprocess',
                                 'process', 'precompletion',
                                 'postprocess']:
                # None of these functions should have been called because we
                # killed the process before they were called
                assert job.result_data != {'test': func_not_run}, \
                    "%s function ran when it shouldn't have" % func_not_run



import os

import requests

from pyqalx.core.entity import QalxEntity
from pyqalx.core.errors import QalxError


class Item(QalxEntity):
    """QalxEntity with entity_type Item"""
    entity_type = 'item'
    _file_bytes = None

    def read_file(self):
        if "file" not in self.keys():
            raise QalxError("Item doesn't have file data.")
        else:
            response = requests.get(url=self['file']['url'])
            if response.ok:
                self._file_bytes = response.content
                return self._file_bytes
            else:
                raise QalxError("Error with file retrieval: \n\n" + response.text)

    def save_file(self, filepath, filename=None):
        if filename is None:
            filename = self['file']['filename']
        if self._file_bytes is None:
            self.read_file()
        with open(os.path.join(filepath, filename), "wb") as f:
            f.write(self._file_bytes)






"""Tests for pyqalx.transport."""

import unittest
import pytest
import os

from requests import exceptions

from unittest.mock import patch, call

from pyqalx.transport.api import PyQalxAPI
from pyqalx.transport.core import PyQalxAPIException


class TestTransportAPI(unittest.TestCase):
    def setUp(self):
        token = '1234567890'
        self.api = PyQalxAPI(config={'TOKEN': token})
        self.headers = {
            'Authorization': 'Token {token}'.format(
                token=token
            )
        }
        self.url = 'https://api.qalx.io'

    def test_base_url_override(self):
        """
        Tests that if the config specifies `BASE_URL` that the pyqalxapi
        base_url is changed. This makes it easier for testing
        """
        api = PyQalxAPI(config={'TOKEN': '12345'})
        assert api.base_url == 'https://api.qalx.io/'

        api = PyQalxAPI(config={'TOKEN': '12345',
                                'BASE_URL': 'http://changed.com'})
        assert api.base_url == 'http://changed.com'

    @patch('requests.request')
    def test_list_item(self, mocked_request):
        """
        Tests listing items including a limit, skip and data.
        This also tests that if the user supplies `params` that
        they are sent off as `json`
        """
        params = {
            'limit': 2,
            'skip': 1,
            'data': ['nested=key=2', 'another=nested=3'],
        }
        self.api.get('item',
                     json=params)
        self.once_with = mocked_request.assert_called_once_with(
            url=self.url + '/item',
            method='GET',
            params=params,
            headers=self.headers)

    @patch('requests.request')
    def test_get_item(self, mocked_request):
        """
        Tests getting a specific item
        """
        self.api.get('item/494b775f-7c95-437b-9989-1c9b730ea67e')
        mocked_request.assert_called_once_with(
            url=self.url + '/item/494b775f-7c95-437b-9989-1c9b730ea67e',
            method='GET',
            headers=self.headers
        )

    @patch('requests.request')
    def test_get_item_params_and_json(self, mocked_request):
        """
        Tests that an exception is raised if `json` and `params`
        are supplied to `get`
        """
        with pytest.raises(PyQalxAPIException):
            self.api.get('item/494b775f-7c95-437b-9989-1c9b730ea67e',
                         params={'some': 'params'},
                         json={'json': 'params'})
        assert mocked_request.called is False

    @patch('requests.request')
    def test_post_item(self, mocked_request):
        """
        Tests creating an item
        """
        self.api.post('item',
                      json={'data': {'a': 'b'},
                            'meta': {'some': 'meta'}},)
        mocked_request.assert_called_once_with(
            url=self.url + '/item',
            method='POST',
            json={'data': {'a': 'b'},
                  'meta': {'some': 'meta'}},
            headers=self.headers
        )

    @patch('requests.request')
    def test_post_file_item_success(self, mocked_request):
        """
        Tests creating a file and uploading it to S3
        """
        mocked_request.return_value.ok = True
        mocked_request.return_value.json.return_value = {
            'file': {
                'put_url': 'https://s3-signed-url.com'
            }
        }

        file_path = os.path.abspath(__file__)
        file_name = os.path.basename(file_path)
        resp = self.api.post('item',
                             json={'data': {'a': 'b'}},
                             file_path=file_path)
        assert 'put_url' not in resp[1]['file'].keys()
        assert mocked_request.call_count == 2
        # First call to create the item
        call1 = call(
            url=self.url + '/item',
            method='POST',
            json={'data': {'a': 'b'},
                  'file': {
                      'name': file_name}},
            headers=self.headers
        )
        # Second call is getting the response
        call2 = call1.json()
        # Third call to PUT the file on S3
        call3 = call(
            url='https://s3-signed-url.com',
            method='PUT',
            data=open(file_path, 'rb').read(),
        )
        calls = [
            call1, call2, call3
        ]
        mocked_request.assert_has_calls(calls)

    @patch('requests.request')
    def test_post_file_item_success_no_upload(self, mocked_request):
        """
        Tests creating a file with upload=False to ensure we don't
        attempt to upload to S3
        """
        mocked_request.return_value.ok = True
        mocked_request.return_value.json.return_value = {
            'file': {
                'put_url': 'https://s3-signed-url.com'
            }
        }

        file_path = os.path.abspath(__file__)
        file_name = os.path.basename(file_path)
        resp = self.api.post('item',
                             json={'data': {'a': 'b'}},
                             file_path=file_path,
                             upload=False)
        # put_url will be there as the user will have to upload manually
        assert 'put_url' in resp[1]['file'].keys()
        assert mocked_request.call_count == 1
        mocked_request.assert_called_once_with(
            url=self.url + '/item',
            method='POST',
            json={'data': {'a': 'b'},
                  'file': {
                      'name': file_name}},
            headers=self.headers
        )

    @patch('requests.request')
    def test_post_file_item_failure(self, mocked_request):
        """
        Tests that we don't attempt to upload to S3 if the file item
        creation failed
        """
        class FirstCall(object):
            def json(self):
                return {
                    'file': {
                        'put_url': 'https://s3-signed-url.com'
                    }
                }

            @property
            def ok(self):
                return True

            @property
            def status_code(self):
                return 200

        class SecondCall(object):
            def json(self):
                return ['some-errors']

            @property
            def ok(self):
                return False

            @property
            def status_code(self):
                return 404

            @property
            def reason(self):
                return 'Bad Request'

        mocked_request.side_effect = [FirstCall(),
                                      SecondCall()]

        file_path = os.path.abspath(__file__)

        resp = self.api.post('item',
                             json={'data': {'a': 'b'}},
                             file_path=file_path)
        expected_data = {
            'status_code': SecondCall().status_code,
            'reason': SecondCall().reason,
            'errors': SecondCall().json()
        }
        assert resp[0] is False
        assert resp[1] == expected_data
        assert mocked_request.call_count == 2

    @patch('requests.request')
    def test_post_file_invalid_file(self, mocked_request):
        """
        Tests that we raise an exception if an invalid file
        is provided
        """
        mocked_request.return_value.ok = True
        mocked_request.return_value.json.return_value = {
            'file': {
                'put_url': 'https://s3-signed-url.com'
            }
        }

        with pytest.raises(PyQalxAPIException):
            self.api.post('item',
                          json={'data': {'a': 'b'}},
                          file_path='INVALIDFILEPATH')
            assert mocked_request.called is False

    @patch('requests.request')
    def test_patch(self, mocked_request):
        """
        Tests a patch request to a specific endpoint
        """
        self.api.patch(
            'queues/494b775f-7c95-437b-9989-1c9b730ea67e',
            json={'data': {'a': 'b'}})
        mocked_request.assert_called_once_with(
            url=self.url + '/queues/494b775f-7c95-437b-9989-1c9b730ea67e',
            method='PATCH',
            json={'data': {'a': 'b'}},
            headers=self.headers
        )

    @patch('requests.request')
    def test_patch_delete_file(self, mocked_request):
        """
        Tests deleting a file from an item
        """
        mocked_request.return_value.ok = True
        mocked_request.return_value.json.return_value = {
            'data': {
                'some': 'data'
            }
        }

        self.api.patch('item/494b775f-7c95-437b-9989-1c9b730ea67e',
                       json={'data': {'some': 'data'}},
                       delete_file=True)
        assert mocked_request.call_count == 1

        url = self.url + '/item/494b775f-7c95-437b-9989-1c9b730ea67e'
        mocked_request.assert_called_once_with(url=url,
                                               method='PATCH',
                                               json={'data': {'some': 'data'},
                                                     'file': {}},
                                               headers=self.headers)

    def test_patch_delete_file_with_file(self):
        """
        Tests that an exception is raised if a user attempts to delete
        a file while also specifying file data
        """
        with pytest.raises(PyQalxAPIException):
            self.api.patch('item/494b775f-7c95-437b-9989-1c9b730ea67e',
                           json={'data': {'some': 'data'}},
                           file_path='/path/to/file',
                           delete_file=True)

    @patch('requests.request')
    def test_delete(self, mocked_request):
        """
        Tests a delete request to a specific endpoint
        """
        self.api.delete('bot/494b775f-7c95-437b-9989-1c9b730ea67e')
        mocked_request.assert_called_once_with(
            url=self.url + '/bot/494b775f-7c95-437b-9989-1c9b730ea67e',
            method='DELETE',
            headers=self.headers
        )

    @patch('requests.request')
    def test_build_request_error(self, mocked_request):
        """
        Tests that we return the correct error data
        if the response is not OK.
        """
        mocked_request.return_value.ok = False
        mocked_request.return_value.status_code = 400
        mocked_request.return_value.json.return_value = 'some-errors'
        mocked_request.return_value.reason = 'Bad Request'
        url = self.api._get_url('item/')
        method = 'GET'
        is_ok, data = self.api._build_request(url=url,
                                              method=method,)
        expected_data = {
            'status_code': 400,
            'reason': 'Bad Request',
            'errors': 'some-errors'
        }
        assert is_ok is False
        assert data == expected_data

    @patch('requests.request')
    def test_patch_file_item_success(self, mocked_request):
        """
        Tests updating a file item and uploading it to S3
        """
        mocked_request.return_value.ok = True
        mocked_request.return_value.json.return_value = {
            'file': {
                'put_url': 'https://s3-signed-url.com'
            }
        }

        file_path = os.path.abspath(__file__)
        file_name = os.path.basename(file_path)
        resp = self.api.patch('item/494b775f-7c95-437b-9989-1c9b730ea67e',
                              json={'data': {'a': 'b'}},
                              file_path=file_path)
        assert 'put_url' not in resp[1]['file'].keys()
        assert mocked_request.call_count == 2
        # First call to create the item
        call1 = call(
            url=self.url + '/item/494b775f-7c95-437b-9989-1c9b730ea67e',
            method='PATCH',
            json={'data': {'a': 'b'},
                  'file': {
                      'name': file_name}},
            headers=self.headers
        )
        # Second call is getting the response
        call2 = call1.json()
        # Third call to PUT the file on S3
        call3 = call(
            url='https://s3-signed-url.com',
            method='PUT',
            data=open(file_path, 'rb').read(),
        )
        calls = [
            call1, call2, call3
        ]
        mocked_request.assert_has_calls(calls)

    @patch('requests.request')
    def test_put(self, mocked_request):
        """
        Tests a put request to a specific endpoint
        """
        self.api.put(
            'queues/494b775f-7c95-437b-9989-1c9b730ea67e',
            json={'data': {'a': 'b'}})
        mocked_request.assert_called_once_with(
            url=self.url + '/queues/494b775f-7c95-437b-9989-1c9b730ea67e',
            method='PUT',
            json={'data': {'a': 'b'}},
            headers=self.headers
        )

    @patch('requests.request')
    def test_put_file_item_success(self, mocked_request):
        """
        Tests replacing a file item and uploading it to S3
        """
        mocked_request.return_value.ok = True
        mocked_request.return_value.json.return_value = {
            'file': {
                'put_url': 'https://s3-signed-url.com'
            }
        }

        file_path = os.path.abspath(__file__)
        file_name = os.path.basename(file_path)
        resp = self.api.put('item/494b775f-7c95-437b-9989-1c9b730ea67e',
                            json={'data': {'a': 'b'}},
                            file_path=file_path)
        assert 'put_url' not in resp[1]['file'].keys()
        assert mocked_request.call_count == 2
        # First call to create the item
        call1 = call(
            url=self.url + '/item/494b775f-7c95-437b-9989-1c9b730ea67e',
            method='PUT',
            json={'data': {'a': 'b'},
                  'file': {
                      'name': file_name}},
            headers=self.headers
        )
        # Second call is getting the response
        call2 = call1.json()
        # Third call to PUT the file on S3
        call3 = call(
            url='https://s3-signed-url.com',
            method='PUT',
            data=open(file_path, 'rb').read(),
        )
        calls = [
            call1, call2, call3
        ]
        mocked_request.assert_has_calls(calls)

    def test_put_delete_file_with_file(self):
        """
        Tests that an exception is raised if a user attempts to delete
        a file while also specifying file data
        """
        with pytest.raises(PyQalxAPIException):
            self.api.put('item/494b775f-7c95-437b-9989-1c9b730ea67e',
                         json={'data': {'some': 'data'}},
                         file_path='/path/to/file',
                         delete_file=True)

    @patch('requests.request')
    def test_put_delete_file(self, mocked_request):
        """
        Tests deleting a file from an item
        """
        mocked_request.return_value.ok = True
        mocked_request.return_value.json.return_value = {
            'data': {
                'some': 'data'
            }
        }

        self.api.put('item/494b775f-7c95-437b-9989-1c9b730ea67e',
                     json={'data': {'some': 'data'}},
                     delete_file=True)
        assert mocked_request.call_count == 1

        url = self.url + '/item/494b775f-7c95-437b-9989-1c9b730ea67e'
        mocked_request.assert_called_once_with(url=url,
                                               method='PUT',
                                               json={'data': {'some': 'data'},
                                                     'file': {}},
                                               headers=self.headers)

    @patch('requests.request')
    def test_requests_exception(self, mocked_request):
        """
        Tests handling of the response if requests raises an exception
        """
        mocked_request.side_effect = exceptions.ConnectionError()

        resp = self.api.get('item')
        expected_resp = (False, {
            'status_code': '',
            'reason': mocked_request.side_effect,
            'errors': []

        })
        assert mocked_request.call_count == 1
        assert resp == expected_resp



import json
import logging
import multiprocessing
import sys
import threading
import traceback
import os
from logging.config import dictConfig
from logging.handlers import RotatingFileHandler


# The default values to use in configs for logging.
# Override the `get_context` method to change this on a per config
# basis
LOGGING_DEFAULTS = {
    'LEVEL': logging._levelToName[logging.ERROR],
    'CONFIG_PATH': None,
    'LOG_FILE_PATH': os.path.join(os.path.expanduser("~"), '.qalx.log')
}


class MultiProcessingLogHandler(logging.Handler):
    # https://gist.github.com/JesseBuesking/10674086
    def __init__(self, filename, mode, maxBytes, backupCount):
        logging.Handler.__init__(self)

        self._handler = RotatingFileHandler(filename, mode,
                                            maxBytes, backupCount)
        self.queue = multiprocessing.Queue(-1)

        t = threading.Thread(target=self.receive)
        t.daemon = True
        t.start()

    def setFormatter(self, fmt):
        logging.Handler.setFormatter(self, fmt)
        self._handler.setFormatter(fmt)

    def receive(self):
        while True:
            try:
                record = self.queue.get()
                self._handler.emit(record)
            except (KeyboardInterrupt, SystemExit):
                raise
            except EOFError:
                break
            except:
                traceback.print_exc(file=sys.stderr)

    def send(self, s):
        self.queue.put_nowait(s)

    def _format_record(self, record):
        # ensure that exc_info and args have been stringified. Removes any
        # chance of unpickleable things inside and possibly reduces message
        # size sent over the pipe
        if record.args:
            record.msg = record.msg % record.args
            record.args = None
        if record.exc_info:
            dummy = self.format(record)
            record.exc_info = None

        return record

    def emit(self, record):
        try:
            s = self._format_record(record)
            self.send(s)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def close(self):
        self._handler.close()
        logging.Handler.close(self)


def build_default_config(LEVEL, DEFAULT_HANDLER, DEFAULT_MP_HANDLER, LOG_FILE_PATH):
    """
    Builds the default config for pyqalx.

    :param LEVEL: The level from the `LOGGING_LEVEL` variable in the
                  users config
    :param DEFAULT_HANDLER: The default log handler we should use.  If a user
        specifies `LOGGING_CONFIG` in their config then this will be
        `nullhandler` (because we assume they know what they are doing and want
        to override various bits of the config). Otherwise it is `filehandler`
        and will log based on the level specified in LEVEL
    :param DEFAULT_MP_HANDLER: The default multiprocessing log handler we
        should use.  If a user specifies `LOGGING_CONFIG` in
        their config then this will be `nullhandler` (because we assume
        they know what they are doing and want to override various bits of
        the config). Otherwise it is `filehandler` and will log based on the
        level specified in LEVEL
    :param LOG_FILE_PATH: The path to the log file to be used when using
        `filehandler` or `mp_filehandler`
    :return: dict: The default config for pyqalx to use.
    """
    DEFAULT_CONFIG = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'default': {
                'format': '%(asctime)s %(name)-12s %(levelname)-8s '
                          '%(process)d %(message)s',
                },
        },
        'handlers': {
            'nullhandler': {
                'class': 'logging.NullHandler',
                'level': LEVEL,
            },
            'filehandler': {
                'class': 'logging.handlers.RotatingFileHandler',
                'filename': LOG_FILE_PATH,
                # 10 MB
                'maxBytes': 10*1024*1024,
                'backupCount': 10,
                'mode': 'w',
                'formatter': 'default',
                'level': LEVEL
            },
            'mp_filehandler': {
                'class': 'pyqalx.core.log.MultiProcessingLogHandler',
                'filename': LOG_FILE_PATH,
                # 10 MB
                'maxBytes': 10*1024*1024,
                'backupCount': 10,
                'mode': 'w',
                'formatter': 'default',
                'level': LEVEL
            }
        },
        'loggers': {
            'pyqalx.core': {
                'handlers': [DEFAULT_HANDLER],
                'level': LEVEL
            },
            'pyqalx.bot': {
                'handlers': [DEFAULT_HANDLER, DEFAULT_MP_HANDLER],
                'level': LEVEL
            },
            'pyqalx.config': {
                'handlers': [DEFAULT_HANDLER],
                'level': LEVEL
            },
            'pyqalx.transport': {
                'handlers': [DEFAULT_HANDLER],
                'level': LEVEL
            },
        },
    }
    return DEFAULT_CONFIG


def configure_logging(config):
    """
    Given a `Config` instance will attempt to get the configurable logging
    values from the profile - falling back to sensible defaults.
    It then configures the logger with our defaults and then overrides as
    necessary depending on if the user has specified a custom logging config
    :param config: An instance of `Config`
    """
    logging_config_func = dictConfig

    LOGGING_LEVEL = config.get('LOGGING_LEVEL').upper()  # noqa
    LOGGING_CONFIG_PATH = config.get('LOGGING_CONFIG_PATH')
    LOG_FILE_PATH = config.get('LOG_FILE_PATH')

    valid_levels = logging._levelToName.values()

    assert LOGGING_LEVEL in valid_levels, \
        'LOGGING_LEVEL must be one of ' \
        '{VALID_LEVELS}.  ' \
        'Value is {VALUE}'.format(VALID_LEVELS=', '.join(valid_levels),
                                  VALUE=LOGGING_LEVEL)

    LEVEL = getattr(logging, LOGGING_LEVEL.upper())

    # If a user is specifying their own config then we modify the default
    # config to use `nullhandler` as we assume the user knows what they are
    # doing and want to override things.  We use 'filehandler' as an easy
    # way to get logs for users who don't want to have the hassle of
    # configuring a logging config themselves
    DEFAULT_HANDLER = 'filehandler' if LOGGING_CONFIG_PATH is None else 'nullhandler'  # noqa
    DEFAULT_MP_HANDLER = 'mp_filehandler' if LOGGING_CONFIG_PATH is None else 'nullhandler'  # noqa

    logging_config_func(build_default_config(LEVEL=LEVEL,
                                             DEFAULT_HANDLER=DEFAULT_HANDLER,
                                             DEFAULT_MP_HANDLER=DEFAULT_MP_HANDLER,  # noqa
                                             LOG_FILE_PATH=LOG_FILE_PATH))

    if LOGGING_CONFIG_PATH is not None:
        assert os.path.exists(LOGGING_CONFIG_PATH), \
            'LOGGING_CONFIG_PATH path does not exist. ' \
            'Path is {LOGGING_CONFIG_PATH}'.format(
                LOGGING_CONFIG_PATH=LOGGING_CONFIG_PATH)
        # They have specified their own config.  Merge it in with the default
        with open(LOGGING_CONFIG_PATH) as f:
            config = json.load(f)
        logging_config_func(config)



import datetime

from copy import deepcopy
from multiprocessing import Process
from time import sleep
from typing import NamedTuple

import sys

from pyqalx import Qalx
from pyqalx.config import BotConfig
from pyqalx.core.errors import QalxBotInitialisationFailed

# We may need below, we may not. What can I say? Dill waters run deep.
# from pyqalx.vendor import dill
# dill.settings['recurse'] = True
from pyqalx.core.log import LOGGING_DEFAULTS


def do_nowt(*args, **kwargs):
    pass


def do_nowt_right(*args, **kwargs):
    return True


class QalxJob:
    __slots__ = ('qalx', 'entity', 'stop',
                 '_current_step', '_process', '_bot_entity', '_queue_message', '_step_results')

    def __init__(self, bot_process):
        """a Job to process

        :param bot_process:
        :type bot_process BotProcess:
        """
        self._process = bot_process
        self.qalx = bot_process.qalx
        self.stop = False
        self._step_results = []
        self._bot_entity = bot_process.api_entity

    def publish_status(self, status_string="", status_dict=None):
        """update the status of the bot

        :param status_string: short message about the bot status
        :type status_string: str
        :param status_dict: structured status data
        """
        status = {
            'time': datetime.datetime.now().isoformat(),
            'message': status_string
        }
        if status_dict is not None:
            status.update(status_dict)
        self.qalx.update_bot_status(self._bot_entity['guid'], status)

    def add_step_result(self, success=True, result_data=None):
        """indicate success and add info about the current step

        :param success: was the step successful or not
        :type success: bool
        :param result_data: data from the result
        :type result_data: dict
        """
        self._step_results.append(StepResult(success=success, step_name=self._current_step, result_data=result_data))

    @property
    def last_step_result(self):
        if self._step_results:
            return self._step_results[-1]
        else:
            return None

    def save_entity(self):
        """save the entity in the message"""
        self.qalx.save_entity(self.entity)


class StepResult(NamedTuple):
    step_name: str = ""
    success: bool = False
    result_data: dict = {}


class BaseBot:

    def __init__(self):
        super().__init__()

    def get_context(self):
        return dict(
            Q_WAITTIMESECONDS=10,  # wait this long between attempts to get a message from the queue
            Q_MSGBATCHSIZE=1,  # number of messages to get in each request
            Q_WAITTIMEMAXSPREADING=12,  # the maximum wait time is this times the Q_WAITTIMESECONDS,
            SAVE_INTERMITTENT=False,  # save the entity between
            SKIP_FINAL_SAVE=False,  # don't save after post-processing
            KILL_AFTER=None,
            LOGGING_LEVEL=LOGGING_DEFAULTS['LEVEL'],  # What level we should log at by default
            LOGGING_CONFIG_PATH=LOGGING_DEFAULTS['CONFIG_PATH'],  # The path to a custom logging dictConfig
            LOG_FILE_PATH=LOGGING_DEFAULTS['LOG_FILE_PATH']  # The path where log files will be stored
        )


class Bot(BaseBot):

    def __init__(self, bot_name, queue_name, user_profile="default", bot_profile_name="default", skip_ini=False,
                 **kwargs):
        super(Bot, self).__init__()
        self.bot_name = bot_name
        self.queue_name = queue_name
        self.kwargs = kwargs
        self.bot_profile_name = bot_profile_name
        self.skip_ini = skip_ini
        self.user_qalx = Qalx(user_profile, skip_ini=skip_ini)

        # bot lifecycle functions
        self.initialisation_function = do_nowt_right
        self.begin_function = do_nowt
        self.preload_function = do_nowt
        self.onload_function = do_nowt
        self.preprocess_function = do_nowt
        self.process_function = do_nowt
        self.precompletion_function = do_nowt
        self.postprocess_function = do_nowt
        self.onwait_function = do_nowt
        self.ontermination_function = do_nowt

    def _collect_steps(self):
        steps = {}
        for fn in self.__dir__():
            if fn.endswith("_function"):
                steps[fn] = getattr(self, fn)  # serialise functions
                # steps[fn] = dill.dumps(getattr(self, fn))  # serialise functions
        return steps

    def start(self, processes=1):
        if self.initialisation_function:
            init_result = self.initialisation_function(self.user_qalx)
            if not init_result:
                raise QalxBotInitialisationFailed("The initialisation function returned a False-like value.")
        steps = self._collect_steps()
        processors = [
            BotProcess(self.config, steps, self.user_qalx, self.queue_name, self.skip_ini, **self.kwargs) for _ in
            range(processes)
        ]
        [
            p.start() for p in processors
        ]
        [
            p.join() for p in processors
        ]

    def initialisation(self, func):
        """decorator to to register the initialisation

        `initialisation` is executed with no argument
        :param func: function to be executed
        :return:
        """
        self.initialisation_function = deepcopy(func)

    def begin(self, func):
        """decorator to to register the begin

        `begin` is executed with no argument
        :param func: function to be executed
        :return:
        """
        self.begin_function = func

    def preload(self, func):
        """decorator executed before entity is loaded from qalx.

        `preload` is executed with a single qalx.QueueMessage argument.
        :param func: function to be executed
        :return:
        """
        self.preload_function = func

    def onload(self, func):
        """decorator to register the onload

        :param func: function to be executed
        :return:
        """
        self.onload_function = func

    def preprocess(self, func):
        """decorator to register preprocess function

        :param func: function to register
        :return:
        """
        self.preprocess_function = func

    def process(self, func):
        """decorator to register process function

        :param func: function to register
        :return:
        """
        self.process_function = func

    def precompletion(self, func):
        """decorator to register precompletion function

        :param func: function to register
        :return:
        """
        self.precompletion_function = func

    def postprocess(self, func):
        """decorator to register postprocess function

        :param func: function to register
        :return:
        """
        self.postprocess_function = func

    def onwait(self, func):
        """decorator to register onwait function

        :param func: function to register
        :return:
        """
        self.onwait_function = func

    def ontermination(self, func):
        """decorator to register ontermination function

        :param func: function to register
        :return:
        """
        self.ontermination_function = func

    @property
    def config(self):
        conf = BotConfig(self.get_context())
        if not self.skip_ini:
            conf.from_inifile(self.bot_profile_name)
        return conf


class BotProcess(Process):

    def __init__(self, config, steps, user_qalx, queue_name, skip_ini, **kwargs):
        super(BotProcess, self).__init__()
        self.config = config
        self.skip_ini = skip_ini
        self.steps = steps

        # at one point in time we got pickle errors - if we get them again then we might need dill as below:
        # self.steps = {}
        # for f_name, f_dill in steps.items(): # de-serialise functions
        #     self.steps[f_name] = dill.loads(f_dill)

        self.queue_name = queue_name
        if "qalx_class" in kwargs:  # to allow testing with various other configs
            qalx_class = kwargs['qalx_class']
            kwargs.pop('qalx_class')
        else:
            qalx_class = Qalx

        self.kwargs = kwargs
        self.api_entity = user_qalx.add_bot(self.config, meta=kwargs)
        self.config['TOKEN'] = self.api_entity['info']['credentials']['token']
        self.qalx = qalx_class(bot_config=self.config, skip_ini=self.skip_ini)
        self.job = QalxJob(self)

    def _check_alive(self):
        if self.job.stop:
            self.steps['ontermination_function'](self.job)
            sys.exit()

    def _run_functions(self):
        self.job.publish_status("running")
        if self.steps['begin_function']:
            self.job._current_step = "Begin"
            self.steps['begin_function'](self.job)
            self._check_alive()

        # get queue
        self.queue = self.qalx.get_queue_by_name(self.queue_name)
        time_to_wait = int(self.config['Q_WAITTIMESECONDS'])
        max_spreading = int(self.config['Q_WAITTIMEMAXSPREADING'])
        kill_after = self.config['KILL_AFTER']
        spread_factor = 1
        no_message_attempts = 0
        while not self.job.stop:
            q_msgs = self.qalx.get_queue_messages(self.queue)
            if q_msgs:
                spread_factor = 1
                for q_msg in q_msgs:
                    self.job._queue_message = q_msg
                    self._process_functions()
                #TODO: I don't think this does anything?
                no_message_attempts = 0
            else:
                self.job._queue_message = None
                self.job._current_step = 'OnWait'
                if self.steps['onwait_function'] is not None:
                    self.steps['onwait_function'](self.job)
                    self._check_alive()
                no_message_attempts += 1

                sleep(time_to_wait * spread_factor)
                next_spread = spread_factor + 1
                spread_factor = min(max_spreading, next_spread)  # never spread beyond a max factor
            if (kill_after is not None) and (no_message_attempts >= int(kill_after)):
                # TODO: I think we need to set self.job._current_step = 'OnTerminate' here??
                self.job.stop = True
                self._check_alive()

    def _process_functions(self):
        def process_as_far_as_possible(functions):
            for fn, step_name in functions:
                if not self.job.stop and fn:
                    self.job._current_step = step_name
                    fn(self.job)
                    self._check_alive()
                    if self.config.getboolean('SAVE_INTERMITTENT'):
                        self.job.save_entity()
                else:
                    break
            if not self.config.getboolean('SKIP_FINAL_SAVE'):
                self.job.save_entity()
            return self.job

        # PreLoad
        if self.steps['preload_function']:
            self.job._current_step = 'PreLoad'
            self.steps['preload_function'](self.job)
            self._check_alive()

        # Process steps
        process_functions = [
            (self.steps['onload_function'], 'OnLoad'),
            (self.steps['preprocess_function'], 'PreProcess'),
            (self.steps['process_function'], 'Process'),
            (self.steps['precompletion_function'], 'PreCompletion'),
            (self.steps['postprocess_function'], 'PostProcess')
        ]
        if not self.job.stop:
            entity_type = self.job._queue_message.body['entity_type']
            entity_guid = self.job._queue_message.body['entity_guid']
            self.job.entity = self.qalx.get_entity_by_guid(entity_type, entity_guid)
            self.queue.delete_message(self.job._queue_message)  # TODO: build in a heartbeat
            process_as_far_as_possible(process_functions)

    def run(self):
        self._run_functions()



import datetime
from uuid import uuid4

item1_guid = uuid4().hex
item2_guid = uuid4().hex
item3_guid = uuid4().hex
item4_guid = uuid4().hex

set1_guid = uuid4().hex
set2_guid = uuid4().hex

group1_guid = uuid4().hex

info = {"created": datetime.datetime.now().isoformat()}
item_file_post = {
    "data": {"some": "data"},
    "file": "/path/to/a/file"
}
item_file_response = {
    "guid": item1_guid,
    "info": info,
    "data": {"some": "data"},
    "file": {
        'url': 'https://s3-signed-url.com'
    }
}
bot_post = {
    'guid': uuid4().hex,
    'info': {
        'credentials': {'token': 'AFAKETOKEN'}
    }
}
queue_post = {
    "parameters": {"some": "param"},
    "meta": {"queue_name": "reeeally unique name..."}
}
queue_response = {
    "guid": uuid4().hex,
    "info": {"created": datetime.datetime.now().isoformat(),
             "queue_url": "https://a.queue.com/123456789",
             "credentials":
                 {
                     'ACCESS_KEY_ID': 'AccessKeyId',
                     'SECRET_ACCESS_KEY': 'SecretAccessKey',
                     'SESSION_TOKEN': 'SessionToken',
                     'EXPIRATION': 'Expiration',
                 }
             },
    "parameters": {"some": "param"},
    "meta": {"queue_name": "reeeally unique name..."}

}
entities_response = {
    'item': [
        {
            "guid": item1_guid,
            "info": info,
            "data": {"some": "data"}
        },
        {
            "guid": item2_guid,
            "info": info,
            "data": {"some": "data"}
        },
        {
            "guid": item3_guid,
            "info": info,
            "data": {"some": "data"}
        },
        {
            "guid": item4_guid,
            "info": info,
            "data": {"some": "data"}
        },
    ],
    "set": [
        {
            "guid": set1_guid,
            "info": info,
            "item_guids": [item1_guid, item2_guid, item3_guid, item4_guid]
        },
        {
            "guid": set2_guid,
            "info": info,
            "item_guids": [item1_guid, item2_guid, item4_guid]
        },
    ],
    "group": [
        {
            "guid": group1_guid,
            "info": info,
            "set_guids": [set1_guid, set2_guid]
        }
    ]
}



"""
.. autoclass:: Qalx
    :members:
"""
import json
import platform
import sys
import logging
from inspect import getmembers, getmro, isclass

from pyqalx.config import UserConfig, BotConfig
from pyqalx.core import QalxNoGUIDError
from pyqalx.core.bot import Bot
from pyqalx.core.entity import QalxEntity
from pyqalx.core.errors import QalxError, \
    QalxEntityNotFound, QalxMultipleEntityReturned, \
    QalxConfigError, QalxAPIResponseError, QalxNoQueueFoundByName
from pyqalx.core.group import Group
from pyqalx.core.item import Item
from pyqalx.core.queue import Queue
from pyqalx.core.set import Set
from pyqalx.transport import PyQalxAPI
from pyqalx.core.log import configure_logging, LOGGING_DEFAULTS

logger = logging.getLogger(__name__)


def entity_by_string(entity_type_name):
    def is_named_entity(o):
        if isclass(o):
            is_qalx_entity = QalxEntity in getmro(o)
            name_matches = getattr(o, "entity_type", None) == entity_type_name
            return is_qalx_entity and (name_matches)
        else:
            return False

    entity_classes = getmembers(sys.modules[__name__], is_named_entity)
    if len(entity_classes) == 0:
        raise QalxEntityNotFound(entity_type_name + " not found.")
    elif len(entity_classes) >= 2:
        raise QalxMultipleEntityReturned(entity_type_name + " matches more than 1 class.")
    else:
        return entity_classes[0][1]


class BaseQalx:

    def __init__(self):
        super().__init__()

    def get_context(self):
        return dict(
            TOKEN="",
            QALX_API_FAIL_SILENT=False,  # if True then API calls which throw HTTP errors won't raise exceptions
            MSG_WAITTIMESECONDS=20,  # calls to the remote queue will not return for this long
            MSG_BLACKOUTSECONDS=30,  # messages must be removed from the queue within this time after reading
            LOGGING_LEVEL=LOGGING_DEFAULTS['LEVEL'], # What level we should log at by default
            LOGGING_CONFIG_PATH=LOGGING_DEFAULTS['CONFIG_PATH'], # The path to a custom logging dictConfig
            LOG_FILE_PATH=LOGGING_DEFAULTS['LOG_FILE_PATH'] # The path where log files will be stored
        )


class Qalx(BaseQalx):

    def __init__(self, profile_name="default", bot_config=None, skip_ini=False):
        """main interface for operations against the qalx REST API

        :param profile_name: user profile name (default="default")
        :type profile_name: basestring
        :param bot_config: a bot config if authenticating as a bot
        :type bot_config: pyqalx.config.BotConfig
        """
        super(Qalx, self).__init__()
        self.config = UserConfig(defaults=self.get_context())
        if not skip_ini:
            self.config.from_inifile(profile_name)
        if bot_config is None:
            self.bot_config = {}
            config = self.config
        elif isinstance(bot_config, BotConfig):
            self.bot_config = bot_config
            config = self.bot_config
        else:
            raise QalxConfigError("A bot_config was supplied but it wasn't a pyqalx.config.BotConfig")

        configure_logging(config)
        self.rest_api = PyQalxAPI(config)

    def _process_api_request(self, method, *args, **kwargs):
        """calls to pyqalxapi

        :param method: http method required
        :param args: args to be passed to endpoint method
        :param kwargs: kwargs to be passed to endpoint method
        :returns: `dict` containing API resource data
        """
        if kwargs.get('meta', None) is None:
            # meta is optional but must always be sent through as a dict
            kwargs['meta'] = {}

        try:
            json.dumps(kwargs)
        except (TypeError, OverflowError):
            raise QalxError("One of the keyword arguments is not JSON serializable")

        try:
            endpoint = getattr(self.rest_api, method.lower())
        except AttributeError:
            raise Exception(f"{method} not recognised.")

        if kwargs.get("file_path"):
            file_path = kwargs.pop("file_path")
            success, data = endpoint(*args, file_path=file_path, json=kwargs)
        else:
            success, data = endpoint(*args, json=kwargs)

        if success:
            return data
        elif not self.config.getboolean("QALX_API_FAIL_SILENT"):
            m = "API request error, message:\n\n-vvv-\n\n"
            m += "\n".join([f"{k}: {v}" for k, v in data.items()])
            m += "\n\n-^^^-"
            raise QalxAPIResponseError(m)
        else:
            pass

    def get_entity_by_guid(self, entity_type, entity_guid):
        """single entity returned by its guid

        :param entity_type: name of entity type e.g. `"set"`
        :type entity_type: str
        :param entity_guid: guid of entity
        :type entity_guid: str
        :type entity_guid: uuid.UUID
        :return: a pyqalx.core.entity.QalxEntity
        """
        endpoint = f"{entity_type}/{entity_guid}"
        rest_api_get = self._process_api_request('get', endpoint)
        class_by_string = entity_by_string(entity_type)
        return class_by_string(
            rest_api_get
        )

    def archive_entity_by_guid(self, entity_type, entity_guid):
        """single entity marked as archived by its guid

        :param entity_type: name of entity type e.g. `"set"`
        :type entity_type: str
        :param entity_guid: guid of entity
        :type entity_guid: str
        :type entity_guid: uuid.UUID
        :return: a pyqalx.core.entity.QalxEntity
        """
        endpoint = f"{entity_type}/{entity_guid}/archive"
        rest_api_get = self._process_api_request('patch', endpoint)
        class_by_string = entity_by_string(entity_type)
        return class_by_string(
            rest_api_get
        )

    def save_entity(self, entity):
        """save changes to a QalxEntity

        :param entity: the entity to save
        :type entity_type: pyqalx.core.entity.QalxEntity

        """
        if not entity.get("guid"):
            raise QalxNoGUIDError("No guid.")
        endpoint = f"{entity.entity_type}/{entity['guid']}"

        keys_to_save = {k: entity[k] for k in entity if k not in ('info', 'guid')}
        rest_api_patch = self._process_api_request('patch', endpoint, **keys_to_save)
        class_by_string = entity_by_string(entity.entity_type)
        return class_by_string(
            rest_api_patch
        )

    def add_item_data(self, data=None, meta=None):
        """adds a dict of data to qalx with optional metadata

        :param data: dictionary of data
        :param meta: (optional) additional data about the item
        :return: pyqalx.core.Item
        """
        if isinstance(data, dict):
            response = self._process_api_request('post', 'item', data=data, meta=meta)
            item = Item(response)
            return item
        else:
            raise QalxError(
                "Only item data in dict format can be added with this function. To add a file try add_item_file ")

    def add_item_file(self, file_path, data=None, meta=None):
        """adds a file to qalx with optional data and metadata

        :param file: path to file
        :param data: (optional) data associated with the file
        :param meta: (optional) additional data about the item
        :return: pyqalx.core.Item
        """
        if data is None:
            data = {}
        response = self._process_api_request('post', 'item', data=data, file_path=file_path, meta=meta)
        return Item(response)

    def add_set(self, items, meta=None):
        """adds a set to qalx with the specified items

        :param items: a sequence of pyqalx.core.Item
        :param meta: (optional) additional data about the item
        :return: pyqalx.core.Set
        """
        item_guids = [item['guid'] for item in items]
        response = self._process_api_request('post', 'set', item_guids=item_guids, meta=meta)
        return Set(response)

    def add_group(self, sets, meta=None):
        """adds a group to qalx with the specified sets

        :param sets: a sequence of pyqalx.core.Set
        :param meta: (optional) additional data about the item
        :return: pyqalx.core.Group
        """
        set_guids = [dataset['guid'] for dataset in sets]

        response = self._process_api_request('post', 'group', set_guids=set_guids, meta=meta)
        return Group(response)

    @property
    def _queue_params(self):
        return {
            'VisibilityTimeout': int(self.config['MSG_BLACKOUTSECONDS'])
            }

    def add_queue(self, queue_name, meta=None):
        """adds a queue to qalx

        :param queue_name: a string identifying the queue
        :param parameters: (optional) additional configuration options for the queue
        :param meta: (optional) additional data about the item
        :return: pyqalx.core.Queue
        """
        if meta is None:
            meta = {"queue_name": queue_name}
        else:
            meta["queue_name"] = queue_name

        response = self._process_api_request('post', 'queue', parameters=self._queue_params, meta=meta)
        return Queue(response)

    def list_entities(self, entity_name, sort=None, skip=0, limit=25, **kwargs):
        """list qalx entities

        :param entity_name: type of entity to list (item, set, etc.)
        :param query_filter: (optional) meta/info key-values to filter by
        :param sort: (optional) meta/info keys to sort by
        :param skip: (default=0) number of records to skip for pagination
        :param limit: (default=25) maximum number of records to return
        :return: list of pyqalx.core.Queue
        """

        entities = self._process_api_request('get', entity_name, sort=sort, skip=skip, limit=limit, **kwargs)

        return [entity_by_string(entity_name)(e) for e in entities['data']]

    def get_queue_messages(self, queue_entity):
        """

        :param queue_entity: the queue to get messages from
        :type queue_entity: pyqalx.core.queue.Queue
        :return:
        """
        if not self.bot_config:
            raise QalxError("Only bots can get messages from queues.")
        max_msgs = int(self.bot_config.get("Q_MSGBATCHSIZE", 1))
        visibility = int(self.bot_config.get("MSG_BLACKOUTSECONDS", 30))
        waittime = int(self.bot_config.get("MSG_WAITTIMESECONDS", 20))
        message = queue_entity.get_messages(max_msgs, visibility, waittime)
        return message

    def get_queue_by_name(self, name):
        """a single queue by name

        :param name: name of queue
        :return: pyqlax.core.Queue
        :raises: pyqalx.errors.QalxReturnedMultipleError
        """

        queues = self.list_entities('queue', meta=[f"queue_name={name}"])
        if len(queues) > 1:
            queues_str = "\n".join([str(q) for q in queues])
            raise QalxMultipleEntityReturned("Expected one but got {}:\n{}".format(len(queues), queues_str))
        elif queues:
            queue = self.get_entity_by_guid('queue', queues[0]['guid'])
            return queue
        else:
            raise QalxNoQueueFoundByName(name + 'not found')

    def get_or_create_queue(self, queue_name, meta=None):
        """

        :param queue_name:
        :type queue_name: str
        :return: pyqalx.core.Queue
        """
        try:
            return self.get_queue_by_name(queue_name)
        except QalxNoQueueFoundByName:
            return self.add_queue(queue_name, meta)

    @property
    def _host_info(self):
        return {
            "node": platform.node(),
            "platform": platform.platform(),
            # TODO: add more platform and IP address infos
        }

    def add_bot(self, bot_config, meta=None):
        """adds a bot resource

        .. note::
            this should only be called by a Bot during startup not by a user directly.

        :param bot_config: the configuration settings for the bot
        :type bot_config: dict
        :param meta: (optional) additional data about the item
        :return: pyqalx.core.Bot
        """
        response = self._process_api_request('post', 'bot',
                                             config=dict(bot_config),
                                             host=self._host_info,
                                             meta=meta)
        return Bot(response)

    def update_bot_status(self, bot_guid, status):
        """updates the status on a bot

        :param bot_guid: the guid of the bot
        :param status: status information to save
        :type status: dict
        :return:
        """
        endpoint = f"bot/{bot_guid}"
        status = self._process_api_request('patch', endpoint, status=status)




import os
from os import environ
from random import randint, choice
from string import printable, whitespace
from tempfile import mkstemp
from time import sleep
from uuid import uuid4

import pytest

from pyqalx import Qalx, Bot
from pyqalx.bot import BaseBot
from pyqalx.config import UserConfig, BotConfig
from pyqalx.core.errors import QalxAPIResponseError, QalxError
from pyqalx.core.log import LOGGING_DEFAULTS
from pyqalx.transport import PyQalxAPI


def append_short_uuid(prefix, uuid_lenth=6):
    return '_'.join([prefix, uuid4().hex[:uuid_lenth]])


DEV_CONFIGS = {"BASE_URL": environ.get("QALX_TEST_URL", "A FAKE TOKEN"),
               "TOKEN": environ.get("QALX_TEST_TOKEN", "A FAKE TOKEN")}

@pytest.fixture(scope="function")
def qalx():
    q = Qalx(skip_ini=True)
    q.config.update(DEV_CONFIGS)
    q.rest_api = PyQalxAPI(q.config)
    return q


@pytest.fixture(scope="function")
def test_bot_and_queue(qalx):
    class TestBot(Bot):
        def __init__(self,*args, **kwargs):
            super(TestBot, self).__init__(*args, **kwargs)

        def get_context(self):
            return dict(

                Q_WAITTIMESECONDS=10,  # wait this long between attempts to get a message from the queue
                Q_MSGBATCHSIZE=1,  # number of messages to get in each request
                Q_WAITTIMEMAXSPREADING=12,  # the maximum wait time is this times the Q_WAITTIMESECONDS,
                SAVE_INTERMITTENT=True,  # save the entity between
                SKIP_FINAL_SAVE=False,  # don't save after post-processing
                KILL_AFTER=1,
                LOGGING_LEVEL=LOGGING_DEFAULTS['LEVEL'],  # What level we should log at by default
                LOGGING_CONFIG_PATH=LOGGING_DEFAULTS['CONFIG_PATH'], # The path to a custom logging dictConfig
                LOG_FILE_PATH=LOGGING_DEFAULTS['LOG_FILE_PATH']  # The path where log files will be stored
            )

        @property
        def config(self):
            conf = BotConfig(self.get_context())
            if not self.skip_ini:
                conf.from_inifile(self.bot_profile_name)
            conf.update(DEV_CONFIGS)
            return conf
    BOT_NAME = append_short_uuid("TEST_BOT")
    QUEUE_NAME = append_short_uuid("TEST_QUEUE")
    bot = TestBot(BOT_NAME, QUEUE_NAME, skip_ini=True)
    bot.user_qalx = qalx
    return bot, QUEUE_NAME


def test_add_item_data(qalx):
    item = qalx.add_item_data({"here_is": "some data"})
    assert item.get("guid")
    qalx.rest_api.delete(f"item/{item.get('guid')}")
    with pytest.raises(QalxAPIResponseError):
        qalx.get_entity_by_guid("item", item.get("guid"))


def test_item_file(qalx):
    _handle, path = mkstemp()
    os.close(_handle)
    random_string = ''.join([choice(printable+whitespace) for _ in range(1000)])
    with open(path, "w") as f:
        f.write(random_string)
    item = qalx.add_item_file(file_path=path)
    assert item.get("guid")
    archived_item = qalx.archive_entity_by_guid('item', item.get("guid"))
    assert archived_item['info']['archived']
    item_guid = item['guid']
    item_by_guid = qalx.get_entity_by_guid('item', item_guid)
    pass


def f_init(qalx):
    print('initialisation')
    return True


def f_begin(job):
    print('begin')
    job.add_step_result(True, {"test": "functional"})


def f_preload(job):
    print('preload')
    job.add_step_result(True, {"test": "functional"})


def f_onload(job):
    print('onload')
    job.entity['data']['onload'] = True
    job.add_step_result(True, {"test": "functional"})


def f_preprocess(job):
    print('preprocess')
    job.entity['data']['preprocess'] = True
    job.add_step_result(True, {"test": "functional"})


def f_process(job):
    how_long = job.entity['data']['how long?']
    sleep(how_long)
    guid = job.entity['guid']
    print(f'process takes {how_long}s on {guid[-6:]}')
    job.entity['data']['process'] = True
    job.add_step_result(True, {"test": "functional"})


def f_precompletion(job):
    print('precompletion')
    job.entity['data']['precompletion'] = True
    job.add_step_result(True, {"test": "functional"})


def f_postprocess(job):
    print('postprocess')
    job.entity['data']['postprocess'] = True
    job.add_step_result(True, {"test": "functional"})


def f_onwait(job):
    print('onwait')
    job.add_step_result(True, {"test": "functional"})


def f_ontermination(job):
    print('ontermination')
    job.add_step_result(True, {"test": "functional"})


def test_bot_items(qalx, test_bot_and_queue):
    test_bot, queue_name = test_bot_and_queue
    queue = qalx.add_queue(queue_name)
    items = []
    for n in range(10):
        item = qalx.add_item_data({"some_data_to_process": "some data",
                                   "how long?": randint(2, 10)})
        queue.submit_entity(item)
        items.append(item)

    test_bot.initialisation_function = f_init
    test_bot.begin_function = f_begin
    test_bot.preload_function = f_preload
    test_bot.onload_function = f_onload
    test_bot.preprocess_function = f_preprocess
    test_bot.process_function = f_process
    test_bot.precompletion_function = f_precompletion
    test_bot.postprocess_function = f_postprocess
    test_bot.onwait_function = f_onwait
    test_bot.ontermination_function = f_ontermination
    test_bot.start(2)
    life_cycle = ('onload',
                  'preprocess',
                  'process',
                  'precompletion',
                  'postprocess')

    for ui in items:
        updated_item = qalx.get_entity_by_guid('item', ui['guid'])
        for part in life_cycle:
            assert updated_item['data'].get(part)


def beam_calcs(job):
    beam_data_guid, weight_data_guid = job.entity['item_guids']  # unpack our input items
    beam_data = job.qalx.get_entity_by_guid('item', beam_data_guid)['data']
    weight_data = job.qalx.get_entity_by_guid('item', weight_data_guid)['data']

    # define a stress calculation function
    def stress_of_beam_in_bending(beam_length, beam_thickness, beam_weight):
        """calculates bending stress in a simply supported beam
        """
        beam_area = beam_length * beam_thickness
        beam_cog = beam_length / 2
        beam_moment = beam_weight * beam_cog
        return beam_moment / beam_area

    stress = stress_of_beam_in_bending(
        beam_data['length'],
        beam_data['thickness'],
        weight_data['value'])

    stress_item = job.qalx.add_item_data(data={
        "Stress": stress,
        "Spec code": beam_data['CODE']
    })
    job.entity['item_guids'].append(stress_item['guid'])
    job.save_entity()


def test_example_set(qalx,  test_bot_and_queue):
    test_bot, queue_name = test_bot_and_queue
    specs = [
        {"CODE": "B1", "thickness": 5, "length": 15},
        {"CODE": "B2", "thickness": 6, "length": 14},
        {"CODE": "B3", "thickness": 4, "length": 17},
    ]

    items = []
    for shape in specs:
        item = qalx.add_item_data(shape, meta={"shape_source": "beam-spec-123"})
        items.append(item)

    weight = {"value": 10, "units": "kg"}
    weight_item = qalx.add_item_data(weight)

    sets = []
    for item in items:
        set_ = qalx.add_set([item, weight_item])
        sets.append(set_)

    group = qalx.add_group(sets, meta={"batch": "Quickstart docs batch"})
    queue = qalx.get_or_create_queue(queue_name)
    queue.submit_sets_from_group(group)

    test_bot.process_function = beam_calcs
    test_bot.start(1)

    for s in sets:
        stressed_set = qalx.get_entity_by_guid('set', s['guid'])
        assert len(stressed_set['item_guids']) == 3  # will only work if the bot added the stress item



"""
.. automodule:: pyqalx.core.qalx
    :members:


Entities
~~~~~~~~

.. automodule:: pyqalx.core.bot
    :members:

.. automodule:: pyqalx.core.entity
    :members:

.. automodule:: pyqalx.core.group
    :members:

.. automodule:: pyqalx.core.item
    :members:

.. automodule:: pyqalx.core.queue
    :members:

.. automodule:: pyqalx.core.set
    :members:

Logging
~~~~~~~

.. automodule:: pyqalx.core.log
    :members:

Errors
~~~~~~

.. automodule:: pyqalx.core.errors
    :members:



"""

from pyqalx.core.errors import QalxNoGUIDError, QalxNoInfoError
from .qalx import Qalx




from pyqalx.core.entity import QalxEntity


class Set(QalxEntity):
    """QalxEntity with entity_type Set

    """
    entity_type = 'set'


    def add_item(self, item):
        """add an item to a set

        :param item:
        :type item: pyqalx.core.Item
        :return: guid of the item
        """
        self['item_guids'].append(item['guid'])
        return item['guid']




from setuptools import setup

from pyqalx import __version__

with open('README.rst', 'r') as f:
    readme = f.read()

setup(name='pyqalx',
      version=__version__,
      description='High-level interfaces to the qalx API',
      long_description=readme,
      classifiers=[

          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Development Status :: 4 - Beta',
          'Intended Audience :: Manufacturing',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Telecommunications Industry',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Natural Language :: English',
          'Operating System :: OS Independent'

      ],
      url='https://docs.qalx.io',
      keywords='qalx engineering simulation',
      author='AgileTek Engineering',
      author_email='',
      license='GNU',
      packages=['pyqalx', 'pyqalx.core', 'pyqalx.config',
                'pyqalx.bot','pyqalx.transport',
                'tests', 'tests.unittests'],
      install_requires=['requests', 'boto3'],
      tests_require=['pytest-mock'],
      include_package_data=True,
      zip_safe=False)



import os
from tempfile import mkstemp

import pytest

from pyqalx import Qalx


def get_qalx_class():
    """
    Helper function for getting our qalx test class
    :return: TestQalx class
    """
    log_handle, temp_log_path = mkstemp()
    os.close(log_handle)

    class TestQalx(Qalx):

        def get_context(self):
            return dict(
                TOKEN="",  # no token is available by default
                QALX_API_FAIL_SILENT=False,
                # if True then API calls which throw HTTP errors won't raise exceptions
                MSG_WAITTIMESECONDS=20,
                # calls to the remote queue will not return for this long
                MSG_BLACKOUTSECONDS=30,
                # messages must be removed from the queue within this time after reading
                LOG_FILE_PATH=temp_log_path,
                LOGGING_LEVEL="ERROR"
            )

    return TestQalx


@pytest.fixture()
def qalx_class():
    return get_qalx_class()


@pytest.fixture()
def qalx():
    return get_qalx_class()(skip_ini=True)


@pytest.fixture
def fake_environ(mocker):
    mocker.patch.dict('pyqalx.config.os.environ',
                      QALX_BOT_THING="A BOT THING",
                      QALX_USER_THING="A USER THING",
                      SOME_OTHER_THING="SECRET")
    return mocker
